import argparse
import json
import networkx as nx
import numpy as np
import random


def parse_args():
  parser = argparse.ArgumentParser()
  parser.add_argument(
    '-n', '--n-nodes', type=int, help='# of nodes', required=True)
  parser.add_argument(
    '--n-links-per-node', type=int, help='# of links', required=True)
  parser.add_argument(
    '-p', '--p-triangle', type=float, help='Probability of adding a triangle after adding a random edge', required=True)
  parser.add_argument(
    '-k', '--n-seeds', type=int, help='# of seed vertices to be selected', required=True)
  parser.add_argument(
    '--alpha', type=float, help='Scale factor of edge propability during simulation', default=1.0)
  parser.add_argument(
    '--coarsening-threshold', type=int, help='Maximum # of nodes after coarsening', required=True)
  parser.add_argument(
    '--min-approximation-ratio', type=float, help='Mimimum approximation ratio', default=0.95)
  parser.add_argument(
    '--max-approximation-ratio', type=float, help='Maximum approximation ratio', default=1.1)
  parser.add_argument(
    '--layout-epsilon', type=float, help='Ratio of layout blank', default=0.1)
  parser.add_argument(
    '-o', '--output-file', type=str, help='JSON output file', default='')
  args = parser.parse_args()

  # Manual checking
  # (1) 1 <= k <= T <= n
  if not 1 <= args.n_seeds <= args.coarsening_threshold <= args.n_nodes:
    raise Exception(f"1 <= n_seeds ({args.n_seeds} given) <= coarsening_threshold ({args.coarsening_threshold} given) "
                    f"<= |V| ({args.n_nodes} given) expected.")
  # (2) p in [0, 1]
  if not 0.0 <= args.p_triangle <= 1.0:
    raise Exception(f"--p-triangle must fall in range [0, 1], while {args.p_triangle} is given.")
  # (3) eps in [0, 1]
  if not 0.0 <= args.layout_epsilon <= 1.0:
    raise Exception(f"--layout-epsilon must fall in range [0, 1], while {args.layout_epsilon} is given.")
  
  return args


def generate_initial_graph(n: int, m: int, p: float):
  while True:
    # Unidirectional graph
    G = nx.powerlaw_cluster_graph(n, m, p)
    if nx.is_connected(G):
      n = G.number_of_nodes()
      if n <= 20:
        print(f"Result graph:")
        for i in range(n):
          print(f"\tG[{i}] = {list(G[i])}")
      return G


# Assumes the graph G is connected
def coarsen_graph(G: nx.Graph):
  n = G.number_of_nodes()
  matches = [-1] * n
  adopts = [-1] * n

  # Step 1: Random Matching
  for v in range(n):
    if matches[v] != -1:
      continue
    candidates = []
    for u in G[v]:
      if matches[u] == -1:
        candidates.append(u)
    if len(candidates) > 0:
      u = random.choice(candidates)
      matches[u] = v
      matches[v] = u
  
  # Step 2: 3-Brotherly Matching or Adoptive Matching
  for v in range(n):
    if matches[v] != -1:
      continue
    pivot_candidates = []
    for u in G[v]:
      if matches[u] == -1:
        raise Exception(f"Implementation error: Unmatched neighbors {u} and {v}.")
      pivot_candidates.append(u)
    # Corner case: v is isolated
    if len(pivot_candidates) == 0:
      matches[v] = v
      continue
    # Otherwise, moves the focus to a random pivot
    pivot = random.choice(pivot_candidates)
    unmatched_neighbors = []
    for u in G[pivot]:
      if matches[u] == -1:
        unmatched_neighbors.append(u)
    
    s = len(unmatched_neighbors)
    if s <= 0:
      raise Exception("Implementation error: no unmatched neighbors.")
    elif s == 1:
      # 2.1: Adoptive Matching
      r = unmatched_neighbors[0]
      x = matches[pivot]
      ap, ax = adopts[pivot], adopts[x]
      if ap != -1 and ax != -1:
        raise Exception(f"Implementation error: Matched neighbors {pivot} and {x} "
                        f"have both adopted {ap} and {ax} respectively.")
      if ap == -1 and ax == -1:
        adopts[pivot] = r
        matches[r] = pivot # Temporary single-direction matching relationship
      elif ap != -1:
        adopts[pivot] = -1
        matches[r], matches[ap] = ap, r
      else:
        adopts[x] = -1
        matches[r], matches[ax] = ax, r
    else:
      if s % 3 == 0:
        head = 0 # 3, 3, 3...
      elif s % 3 == 1:
        head = 4 # 2, 2, 3, 3, 3...
      else:
        head = 2 # 2, 3, 3, 3...
      # Groups the first by size 2
      for i in range(0, head, 2):
        a, b = unmatched_neighbors[i:i + 2]
        matches[a], matches[b] = b, a
      # Groups the rest by size 3
      for i in range(head, s, 3):
        a, b, c = unmatched_neighbors[i:i + 3]
        matches[a], matches[b], matches[c] = b, c, a # Circular matching notation
  
  # Step 3: Grouping nodes
  res = [-1 for _ in range(n)]
  next_index = -1
  for v in range(n):
    if res[v] != -1:
      continue
    next_index += 1

    if matches[v] == v:
      res[v] = next_index
      continue
    u = matches[v]
    if matches[u] == v:
      res[u] = res[v] = next_index
      for a in [adopts[v], adopts[u]]:
        if a != -1:
          res[a] = next_index
      continue
    x = matches[u]
    res[u] = res[v] = res[x] = next_index
  
  # Step 4: Grouping Links
  links = set()
  for u, v in nx.edges(G):
    y, z = res[u], res[v]
    if y < z:
      links.add((y, z))
    elif y > z:
      links.add((z, y))
  
  # (nC, group_index, links), where nC = # of nodes after coarsening,
  # group_index[v] = The node index to which v is coarsened
  return next_index, res, list(links)


args = parse_args()
# Unidirectional graph
G = generate_initial_graph(args.n_nodes, args.n_links_per_node, args.p_triangle)
coarsened_levels = [G]
coarsened_group_info = []
results = {
  "graphs": [],
  "seeds": {
    "cwim": [],
    "contrast": [],
  },
  "simulation": {
    "cwim": 0,
    "contrast": 0,
  }
}


def get_layout(G):
  # For each node v, (x, y) = layout[v], x, y falls in range (-1, 1)
  layout = nx.fruchterman_reingold_layout(G)
  min_xy = min(map(lambda v: np.min(layout[v]), layout))
  max_xy = max(map(lambda v: np.max(layout[v]), layout))
  print(f"{min_xy = }, {max_xy = }")

  # Transforms from (min, max) to (eps/2, 1 - eps/2)
  A = (1.0 - args.layout_epsilon) / (max_xy - min_xy)
  B = (args.layout_epsilon / 2) - A * min_xy

  for v in layout:
    layout[v] = layout[v] * A + B
  
  min_xy = min(map(lambda v: np.min(layout[v]), layout))
  max_xy = max(map(lambda v: np.max(layout[v]), layout))
  print(f"After transforming: {min_xy = }, {max_xy = }")
  return layout


# Step 1: Coarsening
while True:
  prev = coarsened_levels[-1]
  prev_level = len(coarsened_levels) - 1
  print(f"{prev_level = }")

  layout = get_layout(prev)
  cur_step = {
    'nodes': [],
    'links': []
  }
  for v in prev.nodes:
    cur_step['nodes'].append({
      'id': f"{prev_level}-{v}",
      'x': layout[v][0],
      'y': layout[v][1]
    })
  for u, v in prev.edges:
    cur_step['links'].append({
      'source': f"{prev_level}-{u}",
      'target': f"{prev_level}-{v}"
    })

  print(f"{nx.number_of_nodes(prev) = }")

  if nx.number_of_nodes(prev) > args.coarsening_threshold:
    n_coarsened, groups, coarsened_edges = coarsen_graph(prev)
    print(f"Coarsening done. {n_coarsened = }")
    coarsened_group_info.append((n_coarsened, groups))

    for v in range(nx.number_of_nodes(prev)):
      cur_step['nodes'][v]['to'] = f"{prev_level + 1}-{groups[v]}"

    # Builds the coarsened graph
    H = nx.Graph()
    H.add_nodes_from(range(n_coarsened)) # Ensures that node indices are in ascending order
    H.add_edges_from(coarsened_edges)
    coarsened_levels.append(H)
  
  results['graphs'].append(cur_step)
  if nx.number_of_nodes(prev) <= args.coarsening_threshold:
    break


# Step 2: Seed selection (by max-degree) & Seed expansion
n_levels = len(coarsened_levels)
results['seeds']['cwim'] = [[] for _ in range(n_levels)]


def get_seeds_max_degree(G: nx.Graph, n_seeds: int):
  candidates = list(range(nx.number_of_nodes(G)))
  candidates.sort(key=lambda x: G.degree[x], reverse=True)
  # Takes the max-degree as result
  return candidates[0:n_seeds]


def get_seeds(level: int):
  seeds = results['seeds']['cwim']
  if len(seeds[level]) > 0:
    return seeds[level]

  if level == n_levels - 1:
    H = coarsened_levels[-1]
    res = get_seeds_max_degree(H, args.n_seeds)
  else:
    next_seeds = get_seeds(level + 1)
    print(f"Seed of level {level + 1} = {next_seeds}")
    res = []
    # Expands each seed s respectively
    for s in next_seeds:
      candidates = []
      _, groups = coarsened_group_info[level]
      for v in coarsened_levels[level].nodes():
        if groups[v] == s:
          candidates.append(v)
      print(f"In level {level}: Candidate of seed {s} = {candidates}")
      if len(candidates) <= 0:
        raise Exception("Implementation error with candidates.")
      expanded_s = max(candidates, key=lambda v: coarsened_levels[level].degree[v])
      res.append(expanded_s)

  seeds[level] = [f"{level}-{s}" for s in res]
  return res


expanded_seeds = get_seeds(0)
contrast_seeds = get_seeds_max_degree(G, args.n_seeds)
results['seeds']['contrast'] = [f"0-{s}" for s in contrast_seeds]


def simulate(G: nx.Graph, seeds: list[int], alpha: float = 1.0):
  n = nx.number_of_nodes(G)
  p_in = [1.0 - (1.0 - 1.0 / G.degree[i]) ** alpha for i in range(n)]

  vis = [-1] * n
  for s in seeds:
    vis[s] = 0

  queue: list[tuple[int, int]] = [(0, s) for s in seeds]
  res: list[list[object]] = []
  pos = 0
  while pos < len(queue):
    t, cur = queue[pos]
    pos += 1

    for u in G[cur]:
      if vis[u] == -1 and random.random() < p_in[u]:
        vis[u] = t + 1
        queue.append((t + 1, u))
        
        if t < len(res):
          res[t].append({'source': f"0-{cur}", 'target': f"0-{u}"})
        else:
          res.append([{'source': f"0-{cur}", 'target': f"0-{u}"}])
  
  return len(queue), res


# Step 3: Simulation
while True:
  ne, se = simulate(G, expanded_seeds, args.alpha)
  nc, sc = simulate(G, contrast_seeds, args.alpha)
  # Retries until given approximation ratio is reached.
  r = ne / nc
  if args.min_approximation_ratio <= r <= args.max_approximation_ratio:
    print(f"Approximation ratio = {r} ({ne} vs {nc} traversed), with |V| = {nx.number_of_nodes(G)}.")
    results['simulation']['cwim'] = se
    results['simulation']['contrast'] = sc
    break


json_str = json.dumps(results, indent=2)
if args.output_file == '':
  print(json_str)
else:
  with open(args.output_file, mode='w', newline='\n') as file:
    file.write(json_str)
